/*
##############################################################################
# BibItem.java
#
# Stellt ein BibList Element dar.  Hierbei werden die in der BibList
# vorhandenen Elemente (die BibItems) in einer Hashtable aufbewahrt,
# wobei der Hashschlssel der der Wert des key Element des BibItems
# ist und der Hashwert das ganze BibItem.
#
# Beispiel BibList:
# <biblist>
#     	<bibitem key="ID9641745026230" type="InProceedings">
#       	<year>1997</year>
#		<author>M. Zitterbart and T. Harbaum</author>
#		<title>Efficient Routing Table Lookup for IPv6</title>
# 	</bibitem>
#	<bibitem>
#		<...>
#	</bibitem>
# </biblist>
#
# Jonas Pohlandt 2002
##############################################################################
*/
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServlet;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.OutputKeys;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.w3c.dom.*;

import java.util.*;
import java.io.*;
import java.lang.Integer;


public class BibList {

    // ## private Variablen ##

    // In dieser Hashtable werden die BibItems aufbewahrt
    // Hashkey: Wert des key Atributes des BibItems
    // Hashwert: Das BibItem selber
    private Hashtable bibitems = new Hashtable();
    
    // ErrorHandler der beim parsen von XML verwendet wird
    MySaxErrorHandler mySaxEH;
    // String            pathToDTD = BibSearchRessourceLocator.getXMLPath()
    //                               + "biblist.dtd";

    // ## Konstruktoren ##

    // Leerer Konstruktor
    public BibList() {
    }

    // Erstellt eine BibList aus einem DOM Document
    //
    // Parameter: Das DOM Document aus welchem die BibList erzeugt werden soll.
    public BibList(HttpServlet s, Document document) {

	// Alle BibItem Elemente aus dem DOM Document in eine NodeList
	// schreiben
        NodeList list = document.getElementsByTagName("bibitem");

	// Jedes BibItem mit dem Wert seines key Elements als
	// Hashwert in die Hashtable schreiben
        for (int i = 0; i < list.getLength(); i++) {
            Node    node    = list.item(i);    //i-ter Bibitem Node
            BibItem bibitem = new BibItem(s, node);
            bibitems.put(bibitem.getElement("key"), bibitem);
        }
    }

    // Erstellt eine BibList aus einer XML Datei. Dabei wird zuerst
    // die XML Datei in ein DOM document gewandelt, dann verhlt sich
    // der Konstruktor genauso wie der vorhergehende
    //
    // Parameter: Das XML File aus welchem die BibList erzeugt werden
    // soll.
    public BibList(HttpServlet s, File xmlfile) {

        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(true);

        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.setErrorHandler(mySaxEH);
            Document document = builder.parse(xmlfile);
            
            // Alle BibItem Elemente aus dem DOM Document in eine
            // NodeList schreiben
            NodeList list     = document.getElementsByTagName("bibitem");
		
	    // Jedes BibItem mit dem Wert seines key Elements als
	    // Hashwert in die Hashtable schreiben
            for (int i = 0; i < list.getLength(); i++) {
		// s.log("entering node: " + i);
                Node    node    = list.item(i);    //i-ter Bibitem Node
                BibItem bibitem = new BibItem(s, node);
                bibitems.put(bibitem.getElement("key"), bibitem);
            }
            
        // Mgliche Exceptions fangen.
        } catch (SAXException sxe) {
            Exception x = sxe;
            if (sxe.getException() != null) {
                x = sxe.getException();
            }
            x.printStackTrace();
        } catch (ParserConfigurationException pce) {
              pce.printStackTrace();
        } catch (IOException ioe) {
              ioe.printStackTrace();
        }
    }

    // Die BibList ist leer, falls die Hashtable leer ist.
    public boolean isEmpty() {
        boolean empty = bibitems.isEmpty();
        return empty;
    }

    // Diese Methode erzeugt ein DOM Document, welches der BibList entspricht.
    //
    // Rckgabewert: Das erzeugte DOM Document
    public Document toDocument(HttpServletRequest req) {

        Enumeration            e;
        Document               document = null;
        DocumentBuilderFactory factory  =
            DocumentBuilderFactory.newInstance();

        factory.setValidating(true);

        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.setErrorHandler(mySaxEH);
            document = builder.newDocument();
            Element rootelement = (Element) document.createElement("biblist");
            document.appendChild(rootelement);

	    if (req != null) {
		// Request-Daten einfgen
		Element element = (Element)document.createElement("request");
		Enumeration parameternames = req.getParameterNames();
		while (parameternames.hasMoreElements()) {
		    String parameter = (String)(parameternames.nextElement());
		    String value = req.getParameter(parameter).trim();
		    Element child = (Element)document.createElement(parameter);
		    child.appendChild((Text)document.createTextNode(value));
		    element.appendChild(child);
		}
		rootelement.appendChild(element);
	    }

            e = bibitems.keys();
            while (e.hasMoreElements()) {
                String  key     = (String) e.nextElement();
                BibItem bibitem = (BibItem) bibitems.get(key);

                document.getDocumentElement()
                    .appendChild(document.importNode(bibitem.toNode(), true));
            }
        
        // Mgliche Exceptions fangen
        } catch (ParserConfigurationException pce) {
            pce.printStackTrace();
        }

        return document;
    }

    // Diese Methode erzeugt ein DOM Document, welches der BibList entspricht.
    //
    // Rckgabewert: Das erzeugte DOM Document
    public Document toDocument() {
	return toDocument(null);
    }

    // Methoden fr den direkten Zugriff auf die Hashtable
    public int size() {
        return bibitems.size();
    }

    public void addBibItem(BibItem bibitem) {
        bibitems.put(bibitem.getElement("key"), bibitem);
    }

    public void removeBibItem(String uniqueid) {
        bibitems.remove(uniqueid);
    }

    public BibItem getBibItem(String uniqueid) {
        return (BibItem) bibitems.get(uniqueid);
    }

    // Diese Methode erzeugt eine BibList, welche alle BibItems aus
    // der aktuellen BibList enthlt, welche der Suchanfrage aus dem
    // HTTP Request entsprechen. Verleihdaten werden hierbei nicht
    // bercksichtigt.
    //
    // Parameter: Der die Suchanfrage enthaltende Request
    //
    // Rckgabewert: Die Ergebnis BibList
    public BibList search(HttpServlet s, HttpServletRequest req) {

        BibList     resultbiblist = new BibList();
        Enumeration e;
        e = bibitems.keys();

	// Fr alle BibItems der BibList:
        while (e.hasMoreElements()) {
            String  key     = (String) e.nextElement();
            BibItem bibitem = (BibItem) bibitems.get(key);
		
	    // Stimmt das BibItem mit den Parametern der Suchanfrage
	    // berein, wird es zur Ergebnis BibList hinzugefgt
            if (bibitem.matches(s, req)) {
                resultbiblist.addBibItem(bibitem);
            }
        }

        return resultbiblist;
    }

    // Diese Methode erzeugt eine BibList, welche alle BibItems aus
    // der aktuellen BibList enthlt, welche der Suchanfrage aus dem
    // HTTP Request entsprechen. Auch die Verleihdaten werden hierbei
    // bercksichtigt.
    //
    // Parameter:	req: Der die Suchanfrage enthaltende Request
    //		lendlist: Die die Verleihdaten enthaltende LendList
    //
    // Rckgabewert: Die Ergebnis BibList
    public BibList search(HttpServlet s, HttpServletRequest req, LendList lendlist) {

        BibList     resultbiblist = new BibList();
        LendItem    lenditem;
        Enumeration e;
        e = bibitems.keys();
	
	//Fr alle BibItems der BibList:
        while (e.hasMoreElements()) {
            String  key     = (String) e.nextElement();
            BibItem bibitem = (BibItem) bibitems.get(key);
            
            // Zugehriges LendItem der LendList entnehmen 
            lenditem = lendlist.getLendItem(key);
            
            // BibItem mit VerleihInformation anreichern
            bibitem.addLendInfo(lenditem);

	    // Stimmt das BibItem mit den Parametern der Suchanfrage
	    // berein, wird es zur Ergebnis BibList hinzugefgt
	    // s.log("checking " + bibitem.getElement("title"));
            if (bibitem.matches(s, req)) {
		// s.log("bingo!");
                resultbiblist.addBibItem(bibitem);
            }
        }

        return resultbiblist;
    }

    // Diese Methode schreibt die BibList im XML Format in eine
    // Datei. Dies geschieht, indem die BibList zuerst in ein DOM
    // Document gewandelt wird und dann ber die von
    // javax.xml.transform zur Verfgung gestellten Methoden in ein
    // StreamResult (in diesem Fall ein File) geschrieben wird.
    //
    // Parameter: Das zu speichernde File
    public synchronized void toFile(File file) {

        // Document welches spter in das File transformiert wird.
        Document document;
        try {
            // Das Document mit der BibList fllen.
            document = this.toDocument();

	    // Neuen Transformer per TransformerFactory erstellen.
            TransformerFactory tFactory    = TransformerFactory.newInstance();
            Transformer        transformer = tFactory.newTransformer();
            
            // Transformer konfigurieren. Hier wird angegeben, das
            // spter in der XML Datei <!DOCTYPE biblist SYSTEM
            // "[pathToDTD]"> stehen soll
            // transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
	    //                               pathToDTD);

	    // Source und Result fr den Transformer angeben
            DOMSource    source = new DOMSource(document);
            StreamResult result = new StreamResult(file);

            // Transformieren
            transformer.transform(source, result);
        
        // Mgliche Exceptions fangen
        } catch (TransformerConfigurationException tce) {
            System.out.println("\n** Transformer Factory error");
            System.out.println("   " + tce.getMessage());
            Throwable x = tce;
            if (tce.getException() != null) {
                x = tce.getException();
            }
            x.printStackTrace();
        } catch (TransformerException te) {
            System.out.println("\n** Transformation error");
            System.out.println("   " + te.getMessage());
            Throwable x = te;
            if (te.getException() != null) {
                x = te.getException();
            }
            x.printStackTrace();
        }
    }


    // Diese Methode transformiert die BibList ber ein XSLT
    // Stylesheet. Dies geschieht, indem die BibList zuerst in ein DOM
    // Document gewandelt wird und dann ber die von
    // javax.xml.transform zur Verfgung gestellten Methoden in ein
    // StreamResult (in diesem Fall ein StringWriter) geschrieben
    // wird. Auerdem enthlt das erzeugte Dokument die Formular-Daten
    // des angegeben Requests.
    //
    // Parameter: stylesheet  Das zu verwendenede Stylesheet
    //            req         Der Request
    //
    // Rckgabewert: Das Ergebnis der Transformation als String
    public String transform(File stylesheet, HttpServletRequest req) {
        String       resultstring;
        StringWriter swriter = null;
        Document     document;
        
        try {
            document = this.toDocument(req);

	    // Neuen Transformer per TransformerFactory erzeugen.
            TransformerFactory tFactory    = TransformerFactory.newInstance();
            StreamSource       stylesource = new StreamSource(stylesheet);
            
            // Der neue Transformer benutzt die Transfomierungsregeln
            // aus dem Stylesheet
            Transformer        transformer =
                tFactory.newTransformer(stylesource);
        
            swriter = new StringWriter();
            DOMSource    source = new DOMSource(document);
            StreamResult result = new StreamResult(swriter);

            transformer.transform(source, result);
        
        // Mgliche Exceptions fangen
        } catch (TransformerConfigurationException tce) {
            System.out.println("\n** Transformer Factory error");
            System.out.println("   " + tce.getMessage());
            Throwable x = tce;
            if (tce.getException() != null) {
                x = tce.getException();
            }
            x.printStackTrace();
        } catch (TransformerException te) {
            System.out.println("\n** Transformation error");
            System.out.println("   " + te.getMessage());
            Throwable x = te;
            if (te.getException() != null) {
                x = te.getException();
            }
            x.printStackTrace();
        }
	
	resultstring = swriter.toString();
        return resultstring;
    }

    // Diese Methode transformiert die BibList ber ein XSLT
    // Stylesheet. Dies geschieht, indem die BibList zuerst in ein DOM
    // Document gewandelt wird und dann ber die von
    // javax.xml.transform zur Verfgung gestellten Methoden in ein
    // StreamResult (in diesem Fall ein StringWriter) geschrieben
    // wird.
    //
    // Parameter: Das zu verwendenede Stylesheet
    //
    // Rckgabewert: Das Ergebnis der Transformation als String
    public String transform(File stylesheet) {
	return transform(stylesheet, null);
    }

}
